# Copyright (c) 2022 SQL Statement author, see LICENSE file. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

"""Utility functions for list filters analysing sqlparse.sql.Token."""

from sqlparse.sql import (Statement, Token, Identifier, IdentifierList,
    Function, Parenthesis, Comparison, Where)
import sqlparse.tokens as TType

def is_db_or_tablename(token: Token):
    """Returns true if sqlparse.sql.Token is a name of a database or a table."""
    return (token.ttype == TType.Name
        and isinstance(token.parent, Identifier)
        and (isinstance(token.parent.parent, Statement)
            or (isinstance(token.parent.parent, Function)
                and isinstance(token.parent.parent.parent, Statement)
            )
        )
    )

def iscolumnname(token: Token):
    """Returns true if sqlparse.sql.Token is a name of a column
    except for column names in a WHERE clause."""
    return (token.ttype == TType.Name
        and isinstance(token.parent, Identifier)
        and (
            isinstance(token.parent.parent, (Parenthesis, IdentifierList, Comparison))
            or (isinstance(token.parent.parent, Statement)
                and (isdatatypefollowing(token.parent)
                    or iskeywordpreceding(token.parent, "COLUMN")
                )
            )
        )
        and not isinwhere(token)
    )

def isdatatypefollowing(token: Token):
    """Returns true if sqlparse.sql.Token is followed by a column type definition."""
    statement: Statement = token.parent
    idx = statement.token_index(token)

    nextidx, tkn = statement.token_next(idx=idx)
    return isinteger(tkn)

def findstatement(token: Token, lasttoken: Token):
    """Returns sqlparse.sql.Statement by traversing the parent tree upwards.
    Needed internally by some token filtering functions"""
    if isinstance(token, (Statement)):
        return (token, lasttoken)

    return findstatement(token.parent, token)

def iskeyword(token: Token, value: str):
    """Returns true if sqlparse.sql.Token is a specific keyword indicated by value argument."""
    return token.is_keyword and token.normalized == value

def iskeywordpreceding(token: Token, value: str):
    """Returns true if sqlparse.sql.Token is preceeded by a specific keyword.
    Whitespaces skipped."""
    statement: Statement
    tkn: Token
    statement, tkn = findstatement(token, token)

    if tkn.tokens and len(tkn.tokens) > 0 and iskeyword(tkn.tokens[0], value):
        return True

    idx = statement.token_index(tkn)

    previdx, tkn = statement.token_prev(idx=idx)
    return iskeyword(tkn, value)


def isvarchar(token: Token):
    """Returns true if sqlparse.sql.Token is a VARCHAR type."""
    return (token.ttype == TType.Name
        and isinstance(token.parent, Identifier)
        and isinstance(token.parent.parent, Function)
    )

def isinteger(token: Token):
    """Returns true if sqlparse.sql.Token is an INTEGER type."""
    return token.ttype == TType.Name.Builtin

def iscolumntype(token: Token):
    """Returns true if sqlparse.sql.Token is a column type."""
    return isvarchar(token) or isinteger(token)

def getcolumntype(token: Token):
    """Returns column type as a typle specified by sqlparse.sql.Token.
    Possible values are for integer or varchar including its size."""
    if isinteger(token):
        result = (token.value, None)
    else:
        # extract type and size from f.i. varchar(255)
        value: str = token.parent.parent.value
        result = tuple(value.removesuffix(')').split('('))

    return result

def isvaluein(token: Token, keyword: str):
    """Returns true if sqlparse.sql.Token is a value preceeded by a specific keyword."""
    return (token.ttype in (TType.String.Single, TType.Number.Integer)
        and (isinstance(token.parent, (IdentifierList))
            or (isinstance(token.parent, (Comparison))
                and iskeywordpreceding(token.parent, keyword)
            )
        )
    )

def isdata(token: Token):
    """Returns true if sqlparse.sql.Token is a value in UPDATE SET."""
    return isvaluein(token, "SET")

def isinwhere(token: Token):
    """Returns true if sqlparse.sql.Token is a part of WHERE clause."""
    if isinstance(token, (Where)):
        return True

    if isinstance(token, (Statement)):
        return False

    return isinwhere(token.parent)

def isandor(token: Token):
    """Returns true if sqlparse.sql.Token is considered as AND/OR clause."""
    return (
        (token.is_keyword and token.normalized in ('WHERE', 'AND', 'OR'))
        or token.normalized == "("
    )

def getnexttoken(token: Token):
    """Returns next sqlparse.sql.token in a list. Whitespaces skipped."""
    tokens = list(token.parent.tokens)
    nextidx = tokens.index(token) + 1

    while nextidx < len(tokens) and tokens[nextidx].is_whitespace:
        nextidx = nextidx + 1

    return tokens[nextidx]



# Copyright (c) 2022 SQL Statement author,  see LICENSE file. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

"""SQL entities as namedtuples describing the SQL statement for both DDL and DML actions.

Illustrative structure of SQLTable entity:
    SQLTable
        ├── columns
        │   ├── SQLColumn
        │   └── SQLColumn
        │       ├── name (column name)
        │       ├── action (SQLDDLAction or SQLDMLAction)
        │       ├── type (filled for DDL statements)
        │       ├── size (filled for DDL statements, varchar type)
        │       ├── constraints (filled for DDL statements)
        │       │   ├── SQLConstraintPrimaryKey
        │       │   ├── SQLConstraintUnique
        │       │   └── SQLConstraintNotNull
        │       └── value (filled for DML statements such as INSERT or UPDATE)
        └── where
            ├── SQLAnd
            │   └── filter (Contains either:
                            1. a list with one SQLColumn with name and value. Supported simple equal operator. 
                            2. or the list contains nested condition in parentheses)
            ├── SQLAnd
            └── SQLOr
                └── filter (Contains either:
                            1. a list with one SQLColumn with name and value. Supported simple equal operator. 
                            2. or the list contains nested condition in parentheses)
"""

from collections import namedtuple

SQLEntity = namedtuple('SQLEntity', 'name action',)
SQLDatabase = namedtuple('SQLDatabase', SQLEntity._fields)
SQLTable = namedtuple('SQLTable', SQLEntity._fields + ('columns', 'where',), defaults=(None,))
SQLColumn = namedtuple('SQLColumn', SQLEntity._fields + ('type', 'size', 'constraints', 'value'), 
    defaults=(None, None, None))

SQLConstraint = namedtuple('SQLConstraint', SQLEntity._fields)
SQLConstraintPrimaryKey = namedtuple('SQLConstraintPrimaryKey', SQLConstraint._fields)
SQLConstraintUnique = namedtuple('SQLConstraintUnique', SQLConstraint._fields)
SQLConstraintNotNull = namedtuple('SQLConstraintNotNull', SQLConstraint._fields)
SQLConstraintDefault = namedtuple('SQLConstraintDefault', SQLConstraint._fields + ('value',))

SQLAnd = namedtuple('SQLAnd', 'filter',)
SQLOr = namedtuple('SQLOr', 'filter',)



# Copyright (c) 2022 SQL Statement author, see LICENSE file. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

"""This module contains DDL and DML actions in form of enumerations."""

from enum import Enum

class SQLDDLAction(str, Enum):
    """
    Enum listting SQL DDL actions.
    """

    CREATE = "CREATE"
    ALTER = "ALTER"
    DROP = "DROP"
    ADDCONSTRAINT = "ADDCONSTRAINT"
    DROPCONSTRAINT = "DROPCONSTRAINT"
    ADDCOLUMN = "ADDCOLUMN"
    MODIFYCOLUMN = "MODIFYCOLUMN"
    DROPCOLUMN = "DROPCOLUMN"

class SQLDMLAction(str, Enum):
    """
    Enum listting SQL DML actions.
    """

    SELECT      = "SELECT"
    INSERT      = "INSERT"
    UPDATE      = "UPDATE"
    DELETE      = "DELETE"
    WHERE       = "WHERE"
    WHEREEQUAL  = "WHEREEQUAL"
    WHERELIKE   = "WHERELIKE"



import unittest
from typing import List, Tuple
from src.sqlstatement.sql import SQLEntityFactory
from src.sqlstatement.sqlentities import (SQLDatabase, SQLTable, SQLColumn, SQLConstraint,
    SQLConstraintNotNull, SQLConstraintPrimaryKey, SQLConstraintUnique, SQLAnd, SQLOr)
from src.sqlstatement.sqlactions import SQLDDLAction, SQLDMLAction

class SampleSQL:

    CREATEDB  = "CREATE DATABASE testDB;"
    DROPDB  = "DROP DATABASE testDB;"
    CREATETABLE = """CREATE TABLE Persons (
                    PersonID int,
                    LastName varchar(255),
                    FirstName varchar(255),
                    Address varchar(255),
                    City varchar(255)
                );"""

    CREATETABLEEXPECTED = [
        ("PersonID", "int", None, SQLDDLAction.CREATE, []),
        ("LastName", "varchar", "255", SQLDDLAction.CREATE, []),
        ("FirstName", "varchar", "255", SQLDDLAction.CREATE, []),
        ("Address", "varchar", "255", SQLDDLAction.CREATE, []),
        ("City", "varchar", "255", SQLDDLAction.CREATE, [])
    ]

    CREATETABLECONSTRAINTS = """CREATE TABLE Persons (
                    PersonID int PRIMARY KEY,
                    LastName varchar(255) NOT NULL,
                    FirstName varchar(255) UNIQUE NOT NULL,
                    Address varchar(255),
                    City varchar(255)
                );"""
    CREATETABLECONSTRAINTSEXPECTED = [
        ("PersonID", "int", None, SQLDDLAction.CREATE, [(SQLConstraintPrimaryKey, "primarykey", SQLDDLAction.ADDCONSTRAINT)]),
        ("LastName", "varchar", "255", SQLDDLAction.CREATE, [(SQLConstraintNotNull, "notnull", SQLDDLAction.ADDCONSTRAINT)]),
        ("FirstName", "varchar", "255", SQLDDLAction.CREATE, [(SQLConstraintUnique, "unique", SQLDDLAction.ADDCONSTRAINT), (SQLConstraintNotNull, "notnull", SQLDDLAction.ADDCONSTRAINT)]),
        ("Address", "varchar", "255", SQLDDLAction.CREATE, []),
        ("City", "varchar", "255", SQLDDLAction.CREATE, [])
    ]

    ALTERTABLEADD = "ALTER TABLE Persons ADD DateOfBirth date;"
    ALTERTABLEADDEXPECTED = [
        ("DateOfBirth", "date", None, SQLDDLAction.ADDCOLUMN, [])
    ]

    ALTERTABLEADDWCONSTRAINT = "ALTER TABLE Persons ADD DateOfBirth date NOT NULL UNIQUE, customer_name varchar(50) NOT NULL;"
    ALTERTABLEADDWCONSTRAINTEXPECTED = [
        ("DateOfBirth", "date", None, SQLDDLAction.ADDCOLUMN, [(SQLConstraintNotNull, "notnull", SQLDDLAction.ADDCONSTRAINT), (SQLConstraintUnique, "unique", SQLDDLAction.ADDCONSTRAINT)]),
        ("customer_name", "varchar", "50", SQLDDLAction.ADDCOLUMN, [(SQLConstraintNotNull, "notnull", SQLDDLAction.ADDCONSTRAINT)])
    ]

    ALTERTABLEMODIFY = "ALTER TABLE Persons MODIFY COLUMN DateOfBirth date;"
    ALTERTABLEMODIFYEXPECTED = [
        ("DateOfBirth", "date", None, SQLDDLAction.MODIFYCOLUMN, [])
    ]

    ALTERTABLEDROP = "ALTER TABLE Persons DROP COLUMN DateOfBirth;"
    ALTERTABLEDROPEXPECTED = [
        ("DateOfBirth", None, None, SQLDDLAction.DROPCOLUMN, [])
    ]

    ADDUNIQUE = "ALTER TABLE Persons ADD CONSTRAINT UC_Person UNIQUE (ID,LastName);"
    ADDUNIQUEEXPECTED = [
        ("ID", None, None, SQLDDLAction.ADDCONSTRAINT, [(SQLConstraintUnique, "UC_Person", SQLDDLAction.ADDCONSTRAINT)]),
        ("LastName", None, None, SQLDDLAction.ADDCONSTRAINT, [(SQLConstraintUnique, "UC_Person", SQLDDLAction.ADDCONSTRAINT)])
    ]

    ADDPRIMARYKEY = "ALTER TABLE Persons ADD CONSTRAINT PK_Person PRIMARY KEY (ID,LastName);"
    ADDPRIMARYKEYEXPECTED = [
        ("ID", None, None, SQLDDLAction.ADDCONSTRAINT, [(SQLConstraintPrimaryKey, "PK_Person", SQLDDLAction.ADDCONSTRAINT)]),
        ("LastName", None, None, SQLDDLAction.ADDCONSTRAINT, [(SQLConstraintPrimaryKey, "PK_Person", SQLDDLAction.ADDCONSTRAINT)])
    ]

    DROPCONSTRAINT = "ALTER TABLE Persons DROP CONSTRAINT UC_Person;"
    DROPCONSTRAINTEXPECTED = [
        ("*", None, None, SQLDDLAction.DROPCONSTRAINT, [(SQLConstraint, "UC_Person", SQLDDLAction.DROPCONSTRAINT)])        
    ]

    ADDNOTNULL = "ALTER TABLE Persons MODIFY Age int NOT NULL;"
    ADDNOTNULLEXPECTED = [
        ("Age", "int", None, SQLDDLAction.ADDCONSTRAINT, [(SQLConstraintNotNull, "notnull", SQLDDLAction.ADDCONSTRAINT)])
    ]

    DROPTABLE = "DROP TABLE Persons"

    INSERTINTOWCOLS = "INSERT INTO Customers (CustomerName, ContactName, Address, City, PostalCode, Country) VALUES ('Cardinal', 'Tom B. Erichsen', 'Skagen 21', 'Stavanger', 4006, 'Norway');"
    INSERTINTOWCOLSEXPECTED = [
        ("CustomerName", "Cardinal", SQLDMLAction.INSERT),
        ("ContactName", "Tom B. Erichsen", SQLDMLAction.INSERT),
        ("Address", "Skagen 21", SQLDMLAction.INSERT),
        ("City", "Stavanger", SQLDMLAction.INSERT),
        ("PostalCode", "4006", SQLDMLAction.INSERT),
        ("Country", "Norway", SQLDMLAction.INSERT)
    ]

    INSERTINTONOCOLS = "INSERT INTO Customers VALUES ('Cardinal', 'Tom B. Erichsen', 'Skagen 21', 'Stavanger', 4006, 'Norway');"
    INSERTINTONOCOLSEXPECTED = [
        (None, "Cardinal", SQLDMLAction.INSERT),
        (None, "Tom B. Erichsen", SQLDMLAction.INSERT),
        (None, "Skagen 21", SQLDMLAction.INSERT),
        (None, "Stavanger", SQLDMLAction.INSERT),
        (None, "4006", SQLDMLAction.INSERT),
        (None, "Norway", SQLDMLAction.INSERT)
    ]

    UPDATEONE = "UPDATE Customers SET CustomerName='Cardinal' WHERE Country='Mexico';"
    UPDATEONEEXPECTED = [
        ("CustomerName", "Cardinal", SQLDMLAction.UPDATE),        
    ]

    UPDATEMULTI = "UPDATE Customers SET ContactName='Juan', CustomerName='Cardinal' WHERE Country='Mexico' AND City LIKE '%Monterrey%' AND ( ContactName='Juan' OR ContactName='Isabela' );"
    UPDATEMULTIEXPECTED = [
        ("ContactName", "Juan", SQLDMLAction.UPDATE),
        ("CustomerName", "Cardinal", SQLDMLAction.UPDATE),        
    ]    

    SELECTFROM = "SELECT CustomerName, City FROM Customers WHERE Country='Mexico' AND City LIKE '%Monterrey%' AND ( ContactName='Juan' OR ContactName='Isabela' );"
    SELECTFROMEXPECTED = [
        ("CustomerName", None, None, SQLDMLAction.SELECT, []),
        ("City", None, None, SQLDMLAction.SELECT, [])
    ]

    WHEREEXPECTED = [
        (SQLAnd, [("Country", "Mexico", SQLDMLAction.WHEREEQUAL)]),
        (SQLAnd, [("City", "%Monterrey%", SQLDMLAction.WHERELIKE)]),
        (SQLAnd, [
            (SQLAnd, [("ContactName", "Juan", SQLDMLAction.WHEREEQUAL)]), 
            (SQLOr, [("ContactName", "Isabela", SQLDMLAction.WHEREEQUAL)]),
        ]),
    ]

    DELETEFROM = "DELETE FROM Customers WHERE CustomerName='Alfreds Futterkiste';"
    DELETEWHEREEXPECTED = [
        (SQLAnd, [("CustomerName", "Alfreds Futterkiste", SQLDMLAction.WHEREEQUAL)]),
    ]

class TestSQLParse(unittest.TestCase):

    def test_parsecreatedb(self):
        sqlentity: SQLDatabase = SQLEntityFactory.create_entity(SampleSQL.CREATEDB)

        self.assert_entity(sqlentity, SQLDatabase, "testDB", SQLDDLAction.CREATE)

    def test_parsedropdb(self):
        sqlentity: SQLDatabase = SQLEntityFactory.create_entity(SampleSQL.DROPDB)

        self.assert_entity(sqlentity, SQLDatabase, "testDB", SQLDDLAction.DROP)          

    def test_parsecreatetable(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.CREATETABLE)

        self.assert_entity(sqlentity, SQLTable, "Persons", SQLDDLAction.CREATE)
        self.assert_lists(self.assert_column, sqlentity.columns, SampleSQL.CREATETABLEEXPECTED)

    def test_parsecreatetableconstraints(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.CREATETABLECONSTRAINTS)

        self.assert_entity(sqlentity, SQLTable, "Persons", SQLDDLAction.CREATE)
        self.assert_lists(self.assert_column, sqlentity.columns, SampleSQL.CREATETABLECONSTRAINTSEXPECTED)

    def test_parseaddconstraintunique(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.ADDUNIQUE)

        self.assert_entity(sqlentity, SQLTable, "Persons", SQLDDLAction.ADDCONSTRAINT)
        self.assert_lists(self.assert_column, sqlentity.columns, SampleSQL.ADDUNIQUEEXPECTED)

    def test_parseaddconstraintprimarykey(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.ADDPRIMARYKEY)

        self.assert_entity(sqlentity, SQLTable, "Persons", SQLDDLAction.ADDCONSTRAINT)
        self.assert_lists(self.assert_column, sqlentity.columns, SampleSQL.ADDPRIMARYKEYEXPECTED)
                
    def test_parsedropconstraint(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.DROPCONSTRAINT)

        self.assert_entity(sqlentity, SQLTable, "Persons", SQLDDLAction.DROPCONSTRAINT)
        self.assert_lists(self.assert_column, sqlentity.columns, SampleSQL.DROPCONSTRAINTEXPECTED)

    def test_alteraddconstraintnotnull(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.ADDNOTNULL)

        self.assert_entity(sqlentity, SQLTable, "Persons", SQLDDLAction.ADDCONSTRAINT)
        self.assert_lists(self.assert_column, sqlentity.columns, SampleSQL.ADDNOTNULLEXPECTED)

    def test_altertableadd(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.ALTERTABLEADD)

        self.assert_entity(sqlentity, SQLTable, "Persons", SQLDDLAction.ALTER)
        self.assert_lists(self.assert_column, sqlentity.columns, SampleSQL.ALTERTABLEADDEXPECTED)

    def test_altertableaddwconstraint(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.ALTERTABLEADDWCONSTRAINT)

        self.assert_entity(sqlentity, SQLTable, "Persons", SQLDDLAction.ALTER)
        self.assert_lists(self.assert_column, sqlentity.columns, SampleSQL.ALTERTABLEADDWCONSTRAINTEXPECTED)
        
    def test_altertablemodify(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.ALTERTABLEMODIFY)

        self.assert_entity(sqlentity, SQLTable, "Persons", SQLDDLAction.ALTER)
        self.assert_lists(self.assert_column, sqlentity.columns, SampleSQL.ALTERTABLEMODIFYEXPECTED)

    def test_altertabledrop(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.ALTERTABLEDROP)

        self.assert_entity(sqlentity, SQLTable, "Persons", SQLDDLAction.ALTER)
        self.assert_lists(self.assert_column, sqlentity.columns, SampleSQL.ALTERTABLEDROPEXPECTED)

    def test_insertintowcols(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.INSERTINTOWCOLS)

        self.assert_entity(sqlentity, SQLTable, "Customers", SQLDMLAction.INSERT)
        self.assert_lists(self.assert_columndata, sqlentity.columns, SampleSQL.INSERTINTOWCOLSEXPECTED)

    def test_insertintonocols(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.INSERTINTONOCOLS)

        self.assert_entity(sqlentity, SQLTable, "Customers", SQLDMLAction.INSERT)
        self.assert_lists(self.assert_columndata, sqlentity.columns, SampleSQL.INSERTINTONOCOLSEXPECTED)

    def test_updateone(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.UPDATEONE)

        self.assert_entity(sqlentity, SQLTable, "Customers", SQLDMLAction.UPDATE)
        self.assert_lists(self.assert_columndata, sqlentity.columns, SampleSQL.UPDATEONEEXPECTED)

    def test_updatemulti(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.UPDATEMULTI)

        self.assert_entity(sqlentity, SQLTable, "Customers", SQLDMLAction.UPDATE)
        self.assert_lists(self.assert_columndata, sqlentity.columns, SampleSQL.UPDATEMULTIEXPECTED)

        self.assert_where(sqlentity.where, SampleSQL.WHEREEXPECTED)

    def test_selectfrom(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.SELECTFROM)

        self.assert_entity(sqlentity, SQLTable, "Customers", SQLDMLAction.SELECT)
        self.assert_lists(self.assert_column, sqlentity.columns, SampleSQL.SELECTFROMEXPECTED)

        self.assert_where(sqlentity.where, SampleSQL.WHEREEXPECTED)

    def test_deletefrom(self):
        sqlentity: SQLTable = SQLEntityFactory.create_entity(SampleSQL.DELETEFROM)

        self.assert_entity(sqlentity, SQLTable, "Customers", SQLDMLAction.DELETE)        

        self.assert_where(sqlentity.where, SampleSQL.DELETEWHEREEXPECTED)

    def assert_entity(self, sqlentity, enttype:type, name: str, action: SQLDDLAction):
        self.assertIsInstance(sqlentity, enttype)
        self.assertEqual(sqlentity.name, name, 'Name check failed.')
        self.assertEqual(sqlentity.action, action, 'Entity action check failed.')

    def assert_column(self, actual: SQLColumn, expected: Tuple):
        self.assertEqual(actual.name, expected[0], 'Column name check failed.')
        self.assertEqual(actual.type, expected[1], 'Column type check failed.')
        self.assertEqual(actual.size, expected[2], 'Column size check failed.')
        self.assertEqual(actual.action, expected[3], 'Column action check failed.')

        self.assert_lists(self.assert_constraint, actual.constraints, expected[4])

    def assert_columndata(self, actual: SQLColumn, expected: Tuple):
        self.assertEqual(actual.name, expected[0], 'Column name check failed.')
        self.assertEqual(actual.value, expected[1], 'Column value check failed.')        
        self.assertEqual(actual.action, expected[2], 'Column action check failed.')

    def assert_lists(self, assertfunc, actual_result: List, expected_result: List):
        self.assertEqual(len(actual_result), len(expected_result), 'List count check failed.')

        zipped = zip(actual_result, expected_result)
        for actual, expected in zipped:
            assertfunc(actual, expected)

    def assert_constraint(self, actual, expected: Tuple):
        self.assertIsInstance(actual, expected[0])
        self.assertEqual(actual.name, expected[1], 'Constraint name check failed.')
        self.assertEqual(actual.action, expected[2], 'Action check failed.')

    def assert_filter(self, actual, expected: Tuple):
        self.assertIsInstance(actual, expected[0])

        actualfilter = list(filter(lambda i: isinstance(i, (SQLColumn)), actual.filter))
        actualsubfilter = list(filter(lambda i: isinstance(i, (SQLAnd, SQLOr)), actual.filter))
        if actualfilter:
            self.assert_lists(self.assert_columndata, actualfilter, expected[1])
        elif actualsubfilter:
            self.assert_lists(self.assert_filter, actualsubfilter, expected[1])

    def assert_where(self, actual, expected: Tuple):
        self.assert_lists(self.assert_filter, actual, expected)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2022 SQL Statement author, see LICENSE file. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

"""Core of the SQLStatement functionality. SQLEntityFactory handles parsing and 
analysis of the sql statement string."""

from io import UnsupportedOperation
from typing import List
from sqlparse import parse
from sqlparse.sql import Statement, Token, Parenthesis, Comparison, Where
import sqlparse.tokens as TType

from .sqlentities import (SQLDatabase, SQLTable, SQLColumn, SQLConstraint, SQLConstraintUnique, 
    SQLConstraintNotNull, SQLConstraintPrimaryKey, SQLAnd, SQLOr)
from .sqlactions import SQLDDLAction, SQLDMLAction
from . import sqlparseutils


actionmap = {
    "IdentifierList": SQLDMLAction.INSERT,
    "Comparison": SQLDMLAction.UPDATE
}

class SQLEntityFactory:

    """Factory creating structure of SQL entities as metadata based on analysis
    of the sqlparse library output.

    Usage:
    entity: SQLEntity = SQLEntityFactory.create_entity("<SQL statement string>")

    The output is either SQLDatabase or SQLTable. Please see doc string in module sqlentities
    for more details about the structure.
    """

    @classmethod
    def __normalize_funcname(cls, funcname: str):

        funcname = funcname.upper()
        if funcname.startswith("CREATETABLE"):
            funcname = "CREATETABLE"
        elif funcname.startswith("ALTERTABLEADD") and not funcname in [
                "ALTERTABLEADDCONSTRAINTUNIQUE",
                "ALTERTABLEADDCONSTRAINTPRIMARYKEY"]:
            funcname = "ALTERTABLEADD"

        return funcname

    @classmethod
    def create_entity(cls, sql: str):
        """Creates SQLDatabase or SQLTable by analysis of provided SQL string.
        Please see doc string in module sqlentities
        for more details about the structure."""
        statement: Statement = parse(sql)[0]
        keywords = list(map(lambda token: token.value, 
                filter(lambda token: token.is_keyword, statement.tokens)
            )
        )
        funcname = cls.__normalize_funcname("".join(keywords).replace(' ', '').upper())

        return SQLProcessor[funcname](statement)

    @classmethod
    def __getnamesfrom(cls, func, sql: Statement):

        return list(
            map(lambda token: token.value,
                list(filter(func, sql.flatten()))
            )
        )

    @classmethod
    def __gettypesfrom(cls, func, sql: Statement):

        return list(
            map(sqlparseutils.getcolumntype,
                list(filter(func, sql.flatten()))
            )
        )

    @classmethod
    def __getvaluesfrom(cls, func, sql: Statement):

        return list(
            map(lambda token: token.value,
                list(filter(func, sql.flatten()))
            )
        )

    @classmethod
    def __getwhereconditionoperator(cls, token: Token):

        operators = list(filter(lambda t: t.ttype == TType.Operator.Comparison, token.tokens))
        if operators:
            match operators[0].normalized:
                case "=":
                    return SQLDMLAction.WHEREEQUAL
                case "LIKE":
                    return SQLDMLAction.WHERELIKE
                case _:
                    raise UnsupportedOperation()

        return None

    @classmethod
    def __mapwherecondition(cls, andortoken: Token):

        token: Token = sqlparseutils.getnexttoken(andortoken)
        if isinstance(token, (Comparison)):
            condition=[SQLColumn(name=token.left.value,
                action=cls.__getwhereconditionoperator(token),
                type=None, size=None, constraints=None, value=str(token.right.value).strip("'"))
            ]
        elif isinstance(token, (Parenthesis)):
            condition=cls.__getfilterconditions(token.tokens)

        match andortoken.normalized:
            case "WHERE":
                return SQLAnd(filter=condition)
            case "AND":
                return SQLAnd(filter=condition)
            case "(":
                return SQLAnd(filter=condition)
            case "OR":
                return SQLOr(filter=condition)
            case _:
                return (None, None)

    @classmethod
    def __getfilterconditions(cls, tokens: List[Token]):

        return list(
            map(cls.__mapwherecondition,
                list(filter(sqlparseutils.isandor, tokens))
            )
        )

    @classmethod
    def __getactionfrom(cls, func, sql: Statement):        

        return list(
            map(lambda token: actionmap[type(token.parent).__name__],
                list(filter(func, sql.flatten()))
            )
        )

    @classmethod
    def __getwhere(cls, sql: Statement):

        wheretokens: List[Token] = list(filter(lambda token: isinstance(token, (Where)), sql.tokens))
        if wheretokens:
            return cls.__getfilterconditions(wheretokens[0])

        return None

    @classmethod
    def __map_constraints(cls, colname: str, sql: Statement):

        tokens: List = list(sql.flatten())

        def isconstraint(token: Token):
            if token.normalized in ["PRIMARY", "NOT NULL", "UNIQUE"]:
                idx = tokens.index(token)
                subtokens = tokens[:idx-1]
                tkn: Token

                for tkn in reversed(subtokens):
                    if sqlparseutils.iscolumnname(tkn):
                        break

                return tkn.value == colname

            return False       

        def mapconstraint(token: Token):
            match token.normalized:
                case "UNIQUE":
                    return SQLConstraintUnique(name="unique", action=SQLDDLAction.ADDCONSTRAINT)
                case "PRIMARY":
                    return SQLConstraintPrimaryKey(name="primarykey", action=SQLDDLAction.ADDCONSTRAINT)
                case "NOT NULL":
                    return SQLConstraintNotNull(name="notnull", action=SQLDDLAction.ADDCONSTRAINT)
                case _:
                    raise Exception("Unsupported constraint")

        constraints = list(map(mapconstraint, filter(isconstraint, tokens)))
        return constraints

    @classmethod
    def create_sqldatabase(cls, sql: Statement):

        dbname, *_ = cls.__getnamesfrom(sqlparseutils.is_db_or_tablename, sql)
        return SQLDatabase(name=dbname, action=SQLDDLAction.CREATE)

    @classmethod
    def drop_sqldatabase(cls, sql: Statement):

        dbname, *_ = cls.__getnamesfrom(sqlparseutils.is_db_or_tablename, sql)
        return SQLDatabase(name=dbname, action=SQLDDLAction.DROP)

    @classmethod
    def __map_sqltable(cls, sql: Statement, tableaction: SQLDDLAction, columnaction: SQLDDLAction):
        """Extracts and maps the sql data to the SQLStatement data structure"""
        tablename, *_ = cls.__getnamesfrom(sqlparseutils.is_db_or_tablename, sql)
        columnnames = cls.__getnamesfrom(sqlparseutils.iscolumnname, sql)

        types = cls.__gettypesfrom(sqlparseutils.iscolumntype, sql)
        if not types:
            types = [(None, None)] * len(columnnames)

        zipped = list(zip(columnnames, types))

        columns = list(map(
                lambda column: SQLColumn(name=column[0], action=columnaction, type=column[1][0], 
                    size=column[1][1], constraints=cls.__map_constraints(column[0], sql)),
                zipped
            )
        )

        where = cls.__getwhere(sql)

        return SQLTable(name=tablename, action=tableaction, columns=columns, where=where)

    @classmethod
    def create_sqltable(cls, sql: Statement):
        """Analyzes the CREATE TABLE sql statement."""
        return cls.__map_sqltable(sql, SQLDDLAction.CREATE, SQLDDLAction.CREATE)

    @classmethod
    def drop_sqltable(cls, sql: Statement):
        """Analyzes the DROP TABLE sql statement."""
        tablename, *_ = cls.__getnamesfrom(sqlparseutils.is_db_or_tablename, sql)

        return SQLTable(name=tablename, action=SQLDDLAction.DROP, columns=[]) 

    @classmethod
    def alter_sqltableaddcolumn(cls, sql: Statement):
        """Analyzes the ALTER TABLE ADD COLUMN sql statement."""
        return cls.__map_sqltable(sql, SQLDDLAction.ALTER, SQLDDLAction.ADDCOLUMN)

    @classmethod
    def alter_sqltabledropcolumn(cls, sql: Statement):
        """Analyzes the ALTER TABLE DROP COLUMN sql statement."""
        return cls.__map_sqltable(sql, SQLDDLAction.ALTER, SQLDDLAction.DROPCOLUMN)


    @classmethod
    def alter_sqltablemodifycolumn(cls, sql: Statement):
        """Analyzes the ALTER TABLE MODIFY COLUMN sql statement."""
        return cls.__map_sqltable(sql, SQLDDLAction.ALTER, SQLDDLAction.MODIFYCOLUMN)

    @classmethod
    def alter_sqltablemodifynotnull(cls, sql: Statement):
        """Analyzes the ALTER TABLE MODIFY NOT NULL sql statement."""
        tablename, *_ = cls.__getnamesfrom(sqlparseutils.is_db_or_tablename, sql)
        columnnames = cls.__getnamesfrom(sqlparseutils.iscolumnname, sql)

        types = cls.__gettypesfrom(sqlparseutils.iscolumntype, sql)

        zipped = list(zip(columnnames, types))

        columns = list(map(
                lambda column: SQLColumn(name=column[0], action=SQLDDLAction.ADDCONSTRAINT, 
                    type=column[1][0], size=column[1][1], constraints=[
                        SQLConstraintNotNull(name='notnull', action=SQLDDLAction.ADDCONSTRAINT)
                    ]),
                zipped
            )
        )

        return SQLTable(name=tablename, action=SQLDDLAction.ADDCONSTRAINT, columns=columns)


    @classmethod
    def __map_sqltableconstraint(cls, columnnames: List[str], action: SQLDDLAction, constraint):

        return list(map(
                lambda name: SQLColumn(name=name, action=action, 
                    type=None, size=None, constraints=[constraint]),
                columnnames
            )
        )

    @classmethod
    def alter_sqltableaddconstraint(cls, sql: Statement, constrainttype: type):
        """Analyzes the ALTER TABLE ADD CONSTRAINT sql statement."""
        tablename, constraintname = cls.__getnamesfrom(sqlparseutils.is_db_or_tablename, sql)
        columnnames = cls.__getnamesfrom(sqlparseutils.iscolumnname, sql)

        columns = cls.__map_sqltableconstraint(columnnames, SQLDDLAction.ADDCONSTRAINT, 
            constrainttype(name=constraintname, action=SQLDDLAction.ADDCONSTRAINT))
        return SQLTable(name=tablename, action=SQLDDLAction.ADDCONSTRAINT, columns=columns)

    @classmethod
    def alter_sqltableaddconstraintunique(cls, sql: Statement):
        """Analyzes the ALTER TABLE ADD CONSTRAINT UNIQUE sql statement."""
        return cls.alter_sqltableaddconstraint(sql, SQLConstraintUnique)

    @classmethod
    def alter_sqltableaddconstraintprimarykey(cls, sql: Statement):
        """Analyzes the ALTER TABLE ADD CONSTRAINT PRIMARY KEY sql statement."""
        return cls.alter_sqltableaddconstraint(sql, SQLConstraintPrimaryKey)

    @classmethod
    def alter_sqltabledropconstraint(cls, sql: Statement):
        """Analyzes the ALTER TABLE DROP CONSTRAINT sql statement."""
        tablename, constraintname = cls.__getnamesfrom(sqlparseutils.is_db_or_tablename, sql)
        columns = cls.__map_sqltableconstraint(['*'],
            SQLDDLAction.DROPCONSTRAINT, 
            SQLConstraint(name=constraintname, action=SQLDDLAction.DROPCONSTRAINT)
        )

        return SQLTable(name=tablename, action=SQLDDLAction.DROPCONSTRAINT, columns=columns)

    @classmethod
    def __map_sqldata(cls, sql: Statement, tableaction: SQLDMLAction):

        tablename, *_ = cls.__getnamesfrom(sqlparseutils.is_db_or_tablename, sql)
        columnnames = cls.__getnamesfrom(sqlparseutils.iscolumnname, sql)
        data = cls.__getvaluesfrom(sqlparseutils.isdata, sql)
        dataactions = cls.__getactionfrom(sqlparseutils.isdata, sql)

        if not columnnames:
            columnnames = [None] * len(data)        

        zipped = list(zip(columnnames, data, dataactions))

        columns = list(map(
                lambda column: SQLColumn(name=column[0], action=column[2], type=None,
                    size=None, value=str(column[1]).strip("'"), constraints=[]),
                zipped
            )
        )

        where = cls.__getwhere(sql)

        return SQLTable(name=tablename, action=tableaction, columns=columns, where=where)

    @classmethod
    def insert_into_sqltable(cls, sql: Statement):
        """Analyzes the INSERT INTO sql statement."""
        return cls.__map_sqldata(sql, SQLDMLAction.INSERT)

    @classmethod
    def update_sqltable(cls, sql: Statement):
        """Analyzes the UPDATE sql statement."""
        return cls.__map_sqldata(sql, SQLDMLAction.UPDATE)

    @classmethod
    def selectfrom_sqltable(cls, sql: Statement):
        """Analyzes the SELECT FROM sql statement."""
        return cls.__map_sqltable(sql, SQLDMLAction.SELECT, SQLDMLAction.SELECT)

    @classmethod
    def deletefrom_sqltable(cls, sql: Statement):
        """Analyzes the DELETE FROM sql statement."""
        return cls.__map_sqltable(sql, SQLDMLAction.DELETE, SQLDMLAction.DELETE)


SQLProcessor = {
    "CREATEDATABASE": SQLEntityFactory.create_sqldatabase,
    "DROPDATABASE": SQLEntityFactory.drop_sqldatabase,
    "CREATETABLE": SQLEntityFactory.create_sqltable,
    "ALTERTABLEMODIFYCOLUMN": SQLEntityFactory.alter_sqltablemodifycolumn,
    "ALTERTABLEMODIFYNOTNULL": SQLEntityFactory.alter_sqltablemodifynotnull,
    "ALTERTABLEADD": SQLEntityFactory.alter_sqltableaddcolumn,
    "ALTERTABLEADDCONSTRAINTUNIQUE": SQLEntityFactory.alter_sqltableaddconstraintunique,
    "ALTERTABLEADDCONSTRAINTPRIMARYKEY": SQLEntityFactory.alter_sqltableaddconstraintprimarykey,
    "ALTERTABLEDROPCONSTRAINT": SQLEntityFactory.alter_sqltabledropconstraint,
    "ALTERTABLEDROPCOLUMN": SQLEntityFactory.alter_sqltabledropcolumn,
    "DROPTABLE": SQLEntityFactory.drop_sqltable,
    "INSERTINTO": SQLEntityFactory.insert_into_sqltable,
    "UPDATESET": SQLEntityFactory.update_sqltable,
    "UPDATESETWHERE": SQLEntityFactory.update_sqltable,
    "SELECTFROM": SQLEntityFactory.selectfrom_sqltable,
    "DELETEFROM": SQLEntityFactory.deletefrom_sqltable
}
